@@ -33,9 +33,10 @@ use ibc_proto::ibc::core::commitment::v1::MerkleProof;
 use ibc_proto::ibc::core::connection::v1::{
     QueryClientConnectionsRequest, QueryConnectionsRequest,
 };
+use tendermint_rpc::endpoint::broadcast::tx_sync::Response as TxResponse;
 
 use crate::connection::ConnectionMsgType;
-use crate::error::{Error, Kind};
+use crate::error::Error;
 use crate::event::monitor::TxMonitorCmd;
 use crate::keyring::{KeyEntry, KeyRing};
 use crate::light_client::LightClient;
@@ -67,7 +68,7 @@ pub struct QueryPacketOptions {
 }
 
 /// Defines a blockchain as understood by the relayer
-pub trait Chain: Sized {
+pub trait ChainEndpoint: Sized {
     /// Type of light blocks for this chain
     type LightBlock: Send + Sync;
 
@@ -80,12 +81,14 @@ pub trait Chain: Sized {
     /// Type of the client state for this chain
     type ClientState: ClientState;
 
+    type LightClient: LightClient<Self>;
+
     /// Constructs the chain
     fn bootstrap(config: ChainConfig, rt: Arc<TokioRuntime>) -> Result<Self, Error>;
 
     #[allow(clippy::type_complexity)]
     /// Initializes and returns the light client (if any) associated with this chain.
-    fn init_light_client(&self) -> Result<Box<dyn LightClient<Self>>, Error>;
+    fn init_light_client(&self) -> Result<Self::LightClient, Error>;
 
     /// Initializes and returns the event monitor (if any) associated with this chain.
     fn init_event_monitor(
@@ -104,8 +107,19 @@ pub trait Chain: Sized {
     /// Returns the chain's keybase, mutably
     fn keybase_mut(&mut self) -> &mut KeyRing;
 
+    /// Sends one or more transactions with `msgs` to chain and
+    // synchronously wait for it to be committed.
+    fn send_messages_and_wait_commit(
+        &mut self,
+        proto_msgs: Vec<Any>,
+    ) -> Result<Vec<IbcEvent>, Error>;
+
     /// Sends one or more transactions with `msgs` to chain.
-    fn send_msgs(&mut self, proto_msgs: Vec<Any>) -> Result<Vec<IbcEvent>, Error>;
+    /// Non-blocking alternative to `send_messages_and_wait_commit` interface.
+    fn send_messages_and_wait_check_tx(
+        &mut self,
+        proto_msgs: Vec<Any>,
+    ) -> Result<Vec<TxResponse>, Error>;
 
     fn get_signer(&mut self) -> Result<Signer, Error>;
 
@@ -293,7 +307,7 @@ pub trait Chain: Sized {
         trusted_height: ICSHeight,
         target_height: ICSHeight,
         client_state: &AnyClientState,
-        light_client: &mut dyn LightClient<Self>,
+        light_client: &mut Self::LightClient,
     ) -> Result<(Self::Header, Vec<Self::Header>), Error>;
 
     /// Builds the required proofs and the client state for connection handshake messages.
@@ -313,19 +327,19 @@ pub trait Chain: Sized {
                 if !connection_end.state_matches(&State::Init)
                     && !connection_end.state_matches(&State::TryOpen)
                 {
-                    return Err(Kind::ConnOpenTry("bad connection state".to_string()).into());
+                    return Err(Error::bad_connection_state());
                 }
             }
             ConnectionMsgType::OpenAck => {
                 if !connection_end.state_matches(&State::TryOpen)
                     && !connection_end.state_matches(&State::Open)
                 {
-                    return Err(Kind::ConnOpenTry("bad connection state".to_string()).into());
+                    return Err(Error::bad_connection_state());
                 }
             }
             ConnectionMsgType::OpenConfirm => {
                 if !connection_end.state_matches(&State::Open) {
-                    return Err(Kind::ConnOpenTry("bad connection state".to_string()).into());
+                    return Err(Error::bad_connection_state());
                 }
             }
         }
@@ -350,9 +364,7 @@ pub trait Chain: Sized {
                         CommitmentProofBytes::from(consensus_state_proof),
                         client_state_value.latest_height(),
                     )
-                    .map_err(|e| {
-                        Kind::ConnOpenTry("failed to build consensus proof".to_string()).context(e)
-                    })?,
+                    .map_err(Error::consensus_proof)?,
                 );
 
                 client_state = Some(client_state_value);
@@ -369,7 +381,7 @@ pub trait Chain: Sized {
                 None,
                 height.increment(),
             )
-            .map_err(|_| Kind::MalformedProof)?,
+            .map_err(Error::malformed_proof)?,
         ))
     }
 
@@ -384,10 +396,8 @@ pub trait Chain: Sized {
         let channel_proof =
             CommitmentProofBytes::from(self.proven_channel(port_id, channel_id, height)?.1);
 
-        Ok(
-            Proofs::new(channel_proof, None, None, None, height.increment())
-                .map_err(|_| Kind::MalformedProof)?,
-        )
+        Proofs::new(channel_proof, None, None, None, height.increment())
+            .map_err(Error::malformed_proof)
     }
 
     /// Builds the proof for packet messages.
@@ -417,7 +427,7 @@ pub trait Chain: Sized {
             channel_proof,
             height.increment(),
         )
-        .map_err(|_| Kind::MalformedProof)?;
+        .map_err(Error::malformed_proof)?;
 
         Ok((bytes, proofs))
     }
