@@ -43,7 +43,7 @@ use ibc_proto::ibc::core::{
 use crate::{
     config::ChainConfig,
     connection::ConnectionMsgType,
-    error::{Error, Kind},
+    error::Error,
     event::{
         bus::EventBus,
         monitor::{EventBatch, EventReceiver, MonitorCmd, Result as MonitorResult, TxMonitorCmd},
@@ -53,8 +53,8 @@ use crate::{
 };
 
 use super::{
-    handle::{ChainHandle, ChainRequest, ProdChainHandle, ReplyTo, Subscription},
-    Chain,
+    handle::{ChainHandle, ChainRequest, ReplyTo, Subscription},
+    ChainEndpoint,
 };
 
 pub struct Threads {
@@ -62,9 +62,9 @@ pub struct Threads {
     pub event_monitor: Option<thread::JoinHandle<()>>,
 }
 
-pub struct ChainRuntime<C: Chain> {
+pub struct ChainRuntime<Endpoint: ChainEndpoint> {
     /// The specific chain this runtime runs against
-    chain: C,
+    chain: Endpoint,
 
     /// The sender side of a channel to this runtime. Any `ChainHandle` can use this to send
     /// chain requests to this runtime
@@ -84,20 +84,23 @@ pub struct ChainRuntime<C: Chain> {
     tx_monitor_cmd: TxMonitorCmd,
 
     /// A handle to the light client
-    light_client: Box<dyn LightClient<C>>,
+    light_client: Endpoint::LightClient,
 
     #[allow(dead_code)]
     rt: Arc<TokioRuntime>, // Making this future-proof, so we keep the runtime around.
 }
 
-impl<C: Chain + Send + 'static> ChainRuntime<C> {
+impl<Endpoint> ChainRuntime<Endpoint>
+where
+    Endpoint: ChainEndpoint + Send + 'static,
+{
     /// Spawns a new runtime for a specific Chain implementation.
-    pub fn spawn(
+    pub fn spawn<Handle: ChainHandle>(
         config: ChainConfig,
         rt: Arc<TokioRuntime>,
-    ) -> Result<Box<dyn ChainHandle>, Error> {
+    ) -> Result<Handle, Error> {
         // Similar to `from_config`.
-        let chain = C::bootstrap(config, rt.clone())?;
+        let chain = Endpoint::bootstrap(config, rt.clone())?;
 
         // Start the light client
         let light_client = chain.init_light_client()?;
@@ -112,17 +115,17 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     }
 
     /// Initializes a runtime for a given chain, and spawns the associated thread
-    fn init(
-        chain: C,
-        light_client: Box<dyn LightClient<C>>,
+    fn init<Handle: ChainHandle>(
+        chain: Endpoint,
+        light_client: Endpoint::LightClient,
         event_receiver: EventReceiver,
         tx_monitor_cmd: TxMonitorCmd,
         rt: Arc<TokioRuntime>,
-    ) -> (Box<dyn ChainHandle>, thread::JoinHandle<()>) {
+    ) -> (Handle, thread::JoinHandle<()>) {
         let chain_runtime = Self::new(chain, light_client, event_receiver, tx_monitor_cmd, rt);
 
         // Get a handle to the runtime
-        let handle = chain_runtime.handle();
+        let handle: Handle = chain_runtime.handle();
 
         // Spawn the runtime & return
         let id = handle.id();
@@ -137,8 +140,8 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
 
     /// Basic constructor
     fn new(
-        chain: C,
-        light_client: Box<dyn LightClient<C>>,
+        chain: Endpoint,
+        light_client: Endpoint::LightClient,
         event_receiver: EventReceiver,
         tx_monitor_cmd: TxMonitorCmd,
         rt: Arc<TokioRuntime>,
@@ -157,11 +160,11 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
         }
     }
 
-    pub fn handle(&self) -> Box<dyn ChainHandle> {
-        let chain_id = self.chain.id().clone();
+    pub fn handle<Handle: ChainHandle>(&self) -> Handle {
+        let chain_id = ChainEndpoint::id(&self.chain).clone();
         let sender = self.request_sender.clone();
 
-        Box::new(ProdChainHandle::new(chain_id, sender))
+        Handle::new(chain_id, sender)
     }
 
     fn run(mut self) -> Result<(), Error> {
@@ -175,17 +178,19 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
                         },
                         Err(e) => {
                             error!("received error via event bus: {}", e);
-                            return Err(Kind::Channel.into());
+                            return Err(Error::channel_receive(e));
                         },
                     }
                 },
                 recv(self.request_receiver) -> event => {
                     match event {
                         Ok(ChainRequest::Shutdown { reply_to }) => {
-                            self.tx_monitor_cmd.send(MonitorCmd::Shutdown).map_err(Kind::channel)?;
+                            self.tx_monitor_cmd.send(MonitorCmd::Shutdown)
+                                .map_err(Error::send)?;
 
                             let res = self.chain.shutdown();
-                            reply_to.send(res).map_err(Kind::channel)?;
+                            reply_to.send(res)
+                                .map_err(Error::send)?;
 
                             break;
                         }
@@ -194,8 +199,12 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
                             self.subscribe(reply_to)?
                         },
 
-                        Ok(ChainRequest::SendMsgs { proto_msgs, reply_to }) => {
-                            self.send_msgs(proto_msgs, reply_to)?
+                        Ok(ChainRequest::SendMessagesAndWaitCommit { proto_msgs, reply_to }) => {
+                            self.send_messages_and_wait_commit(proto_msgs, reply_to)?
+                        },
+
+                        Ok(ChainRequest::SendMessagesAndWaitCheckTx { proto_msgs, reply_to }) => {
+                            self.send_messages_and_wait_check_tx(proto_msgs, reply_to)?
                         },
 
                         Ok(ChainRequest::Signer { reply_to }) => {
@@ -350,19 +359,31 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn subscribe(&mut self, reply_to: ReplyTo<Subscription>) -> Result<(), Error> {
         let subscription = self.event_bus.subscribe();
 
-        reply_to.send(Ok(subscription)).map_err(Kind::channel)?;
+        reply_to.send(Ok(subscription)).map_err(Error::send)?;
 
         Ok(())
     }
 
-    fn send_msgs(
+    fn send_messages_and_wait_commit(
         &mut self,
         proto_msgs: Vec<prost_types::Any>,
         reply_to: ReplyTo<Vec<IbcEvent>>,
     ) -> Result<(), Error> {
-        let result = self.chain.send_msgs(proto_msgs);
+        let result = self.chain.send_messages_and_wait_commit(proto_msgs);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
+
+        Ok(())
+    }
+
+    fn send_messages_and_wait_check_tx(
+        &mut self,
+        proto_msgs: Vec<prost_types::Any>,
+        reply_to: ReplyTo<Vec<tendermint_rpc::endpoint::broadcast::tx_sync::Response>>,
+    ) -> Result<(), Error> {
+        let result = self.chain.send_messages_and_wait_check_tx(proto_msgs);
+
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -370,7 +391,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn query_latest_height(&self, reply_to: ReplyTo<Height>) -> Result<(), Error> {
         let latest_height = self.chain.query_latest_height();
 
-        reply_to.send(latest_height).map_err(Kind::channel)?;
+        reply_to.send(latest_height).map_err(Error::send)?;
 
         Ok(())
     }
@@ -378,7 +399,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn get_signer(&mut self, reply_to: ReplyTo<Signer>) -> Result<(), Error> {
         let result = self.chain.get_signer();
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -386,7 +407,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn get_key(&mut self, reply_to: ReplyTo<KeyEntry>) -> Result<(), Error> {
         let result = self.chain.get_key();
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -394,7 +415,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn module_version(&self, port_id: PortId, reply_to: ReplyTo<String>) -> Result<(), Error> {
         let result = self.chain.query_module_version(&port_id);
 
-        reply_to.send(Ok(result)).map_err(Kind::channel)?;
+        reply_to.send(Ok(result)).map_err(Error::send)?;
 
         Ok(())
     }
@@ -412,7 +433,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
                 trusted_height,
                 target_height,
                 &client_state,
-                self.light_client.as_mut(),
+                &mut self.light_client,
             )
             .map(|(header, support)| {
                 let header = header.wrap_any();
@@ -420,7 +441,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
                 (header, support)
             });
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -436,7 +457,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .build_client_state(height)
             .map(|cs| cs.wrap_any());
 
-        reply_to.send(client_state).map_err(Kind::channel)?;
+        reply_to.send(client_state).map_err(Error::send)?;
 
         Ok(())
     }
@@ -456,7 +477,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .build_consensus_state(verified.target)
             .map(|cs| cs.wrap_any());
 
-        reply_to.send(consensus_state).map_err(Kind::channel)?;
+        reply_to.send(consensus_state).map_err(Error::send)?;
 
         Ok(())
     }
@@ -472,7 +493,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .light_client
             .check_misbehaviour(update_event, &client_state);
 
-        reply_to.send(misbehaviour).map_err(Kind::channel)?;
+        reply_to.send(misbehaviour).map_err(Error::send)?;
 
         Ok(())
     }
@@ -495,7 +516,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
         let result = result
             .map(|(opt_client_state, proofs)| (opt_client_state.map(|cs| cs.wrap_any()), proofs));
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -507,7 +528,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_clients(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -519,7 +540,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_client_connections(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -535,7 +556,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .query_client_state(&client_id, height)
             .map(|cs| cs.wrap_any());
 
-        reply_to.send(client_state).map_err(Kind::channel)?;
+        reply_to.send(client_state).map_err(Error::send)?;
 
         Ok(())
     }
@@ -550,7 +571,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .query_upgraded_client_state(height)
             .map(|(cl, proof)| (cl.wrap_any(), proof));
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -562,7 +583,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let consensus_states = self.chain.query_consensus_states(request);
 
-        reply_to.send(consensus_states).map_err(Kind::channel)?;
+        reply_to.send(consensus_states).map_err(Error::send)?;
 
         Ok(())
     }
@@ -578,7 +599,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             self.chain
                 .query_consensus_state(client_id, consensus_height, query_height);
 
-        reply_to.send(consensus_state).map_err(Kind::channel)?;
+        reply_to.send(consensus_state).map_err(Error::send)?;
 
         Ok(())
     }
@@ -593,7 +614,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .query_upgraded_consensus_state(height)
             .map(|(cs, proof)| (cs.wrap_any(), proof));
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -601,7 +622,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn query_commitment_prefix(&self, reply_to: ReplyTo<CommitmentPrefix>) -> Result<(), Error> {
         let prefix = self.chain.query_commitment_prefix();
 
-        reply_to.send(prefix).map_err(Kind::channel)?;
+        reply_to.send(prefix).map_err(Error::send)?;
 
         Ok(())
     }
@@ -609,7 +630,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     fn query_compatible_versions(&self, reply_to: ReplyTo<Vec<Version>>) -> Result<(), Error> {
         let versions = self.chain.query_compatible_versions();
 
-        reply_to.send(versions).map_err(Kind::channel)?;
+        reply_to.send(versions).map_err(Error::send)?;
 
         Ok(())
     }
@@ -622,7 +643,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let connection_end = self.chain.query_connection(&connection_id, height);
 
-        reply_to.send(connection_end).map_err(Kind::channel)?;
+        reply_to.send(connection_end).map_err(Error::send)?;
 
         Ok(())
     }
@@ -634,7 +655,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_connections(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -646,9 +667,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_connection_channels(request);
 
-        reply_to
-            .send(result)
-            .map_err(|e| Kind::Channel.context(e))?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -660,7 +679,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_channels(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -674,7 +693,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_channel(&port_id, &channel_id, height);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -686,7 +705,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_channel_client_state(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -702,7 +721,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .proven_client_state(&client_id, height)
             .map(|(cs, mp)| (cs.wrap_any(), mp));
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -715,7 +734,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.proven_connection(&connection_id, height);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -732,7 +751,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .proven_client_consensus(&client_id, consensus_height, height)
             .map(|(cs, mp)| (cs.wrap_any(), mp));
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -748,7 +767,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             .chain
             .build_channel_proofs(&port_id, &channel_id, height);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -766,7 +785,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
             self.chain
                 .build_packet_proofs(packet_type, port_id, channel_id, sequence, height);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -778,7 +797,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_packet_commitments(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -790,7 +809,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_unreceived_packets(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -802,7 +821,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_packet_acknowledgements(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -814,7 +833,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_unreceived_acknowledgements(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -826,7 +845,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_next_sequence_receive(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
@@ -838,7 +857,7 @@ impl<C: Chain + Send + 'static> ChainRuntime<C> {
     ) -> Result<(), Error> {
         let result = self.chain.query_txs(request);
 
-        reply_to.send(result).map_err(Kind::channel)?;
+        reply_to.send(result).map_err(Error::send)?;
 
         Ok(())
     }
