@@ -1,6 +1,7 @@
 //! Relayer configuration
 
 pub mod reload;
+pub mod types;
 
 use std::collections::{HashMap, HashSet};
 use std::{fmt, fs, fs::File, io::Write, path::Path, time::Duration};
@@ -11,7 +12,8 @@ use tendermint_light_client::types::TrustThreshold;
 use ibc::ics24_host::identifier::{ChainId, ChannelId, PortId};
 use ibc::timestamp::ZERO_DURATION;
 
-use crate::error;
+use crate::config::types::{MaxMsgNum, MaxTxSize};
+use crate::error::Error;
 
 #[derive(Clone, Debug, PartialEq, Serialize, Deserialize)]
 pub struct GasPrice {
@@ -108,6 +110,8 @@ pub struct Config {
     #[serde(default)]
     pub global: GlobalConfig,
     #[serde(default)]
+    pub rest: RestConfig,
+    #[serde(default)]
     pub telemetry: TelemetryConfig,
     #[serde(default = "Vec::new", skip_serializing_if = "Vec::is_empty")]
     pub chains: Vec<ChainConfig>,
@@ -240,6 +244,24 @@ impl Default for TelemetryConfig {
     }
 }
 
+#[derive(Clone, Debug, Deserialize, Serialize)]
+#[serde(deny_unknown_fields)]
+pub struct RestConfig {
+    pub enabled: bool,
+    pub host: String,
+    pub port: u16,
+}
+
+impl Default for RestConfig {
+    fn default() -> Self {
+        Self {
+            enabled: false,
+            host: "127.0.0.1".to_string(),
+            port: 3000,
+        }
+    }
+}
+
 #[derive(Clone, Debug, Deserialize, Serialize)]
 #[serde(deny_unknown_fields)]
 pub struct ChainConfig {
@@ -254,8 +276,10 @@ pub struct ChainConfig {
     pub store_prefix: String,
     pub max_gas: Option<u64>,
     pub gas_adjustment: Option<f64>,
-    pub max_msg_num: Option<usize>,
-    pub max_tx_size: Option<usize>,
+    #[serde(default)]
+    pub max_msg_num: MaxMsgNum,
+    #[serde(default)]
+    pub max_tx_size: MaxTxSize,
     #[serde(default = "default::clock_drift", with = "humantime_serde")]
     pub clock_drift: Duration,
     #[serde(default = "default::trusting_period", with = "humantime_serde")]
@@ -270,34 +294,31 @@ pub struct ChainConfig {
 }
 
 /// Attempt to load and parse the TOML config file as a `Config`.
-pub fn load(path: impl AsRef<Path>) -> Result<Config, error::Error> {
-    let config_toml =
-        std::fs::read_to_string(&path).map_err(|e| error::Kind::ConfigIo.context(e))?;
+pub fn load(path: impl AsRef<Path>) -> Result<Config, Error> {
+    let config_toml = std::fs::read_to_string(&path).map_err(Error::config_io)?;
 
-    let config =
-        toml::from_str::<Config>(&config_toml[..]).map_err(|e| error::Kind::Config.context(e))?;
+    let config = toml::from_str::<Config>(&config_toml[..]).map_err(Error::config_decode)?;
 
     Ok(config)
 }
 
 /// Serialize the given `Config` as TOML to the given config file.
-pub fn store(config: &Config, path: impl AsRef<Path>) -> Result<(), error::Error> {
+pub fn store(config: &Config, path: impl AsRef<Path>) -> Result<(), Error> {
     let mut file = if path.as_ref().exists() {
         fs::OpenOptions::new().write(true).truncate(true).open(path)
     } else {
         File::create(path)
     }
-    .map_err(|e| error::Kind::Config.context(e))?;
+    .map_err(Error::config_io)?;
 
     store_writer(config, &mut file)
 }
 
 /// Serialize the given `Config` as TOML to the given writer.
-pub(crate) fn store_writer(config: &Config, mut writer: impl Write) -> Result<(), error::Error> {
-    let toml_config =
-        toml::to_string_pretty(&config).map_err(|e| error::Kind::Config.context(e))?;
+pub(crate) fn store_writer(config: &Config, mut writer: impl Write) -> Result<(), Error> {
+    let toml_config = toml::to_string_pretty(&config).map_err(Error::config_encode)?;
 
-    writeln!(writer, "{}", toml_config).map_err(|e| error::Kind::Config.context(e))?;
+    writeln!(writer, "{}", toml_config).map_err(Error::config_io)?;
 
     Ok(())
 }
