@@ -3,15 +3,17 @@ use std::{fmt, thread, time::Duration};
 
 use itertools::Itertools;
 use prost_types::Any;
-use thiserror::Error;
 use tracing::{debug, error, info, trace, warn};
 
+use crate::error::Error as RelayerError;
+use flex_error::define_error;
 use ibc::downcast;
 use ibc::events::{IbcEvent, IbcEventType};
 use ibc::ics02_client::client_consensus::{
     AnyConsensusState, AnyConsensusStateWithHeight, ConsensusState, QueryClientEventRequest,
 };
 use ibc::ics02_client::client_state::ClientState;
+use ibc::ics02_client::error::Error as ClientError;
 use ibc::ics02_client::events::UpdateClient;
 use ibc::ics02_client::header::Header;
 use ibc::ics02_client::misbehaviour::MisbehaviourEvidence;
@@ -32,50 +34,195 @@ const MAX_MISBEHAVIOUR_CHECK_DURATION: Duration = Duration::from_secs(120);
 
 const MAX_RETRIES: usize = 5;
 
-#[derive(Debug, Error)]
-pub enum ForeignClientError {
-    #[error("error raised while creating client: {0}")]
-    ClientCreate(String),
+define_error! {
+    ForeignClientError {
+        ClientCreate
+            {
+                chain_id: ChainId,
+                description: String
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("error raised while creating client for chain {0}: {1}",
+                    e.chain_id, e.description)
+            },
 
-    #[error("error raised while updating client: {0}")]
-    ClientUpdate(String),
+        Client
+            [ ClientError ]
+            |_| { "ICS02 client error" },
 
-    #[error("error raised while trying to refresh client {0}: {1}")]
-    ClientRefresh(ClientId, String),
+        ClientUpdate
+            {
+                chain_id: ChainId,
+                description: String
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("error raised while updating client on chain {0}: {1}", e.chain_id, e.description)
+            },
 
-    #[error("failed while querying for client {0} on chain id: {1} with error: {2}")]
-    ClientQuery(ClientId, ChainId, String),
+        ClientAlreadyUpToDate
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+                height: Height,
+            }
+            |e| {
+                format_args!("Client {} is already up-to-date with chain {}@{}",
+                    e.client_id, e.chain_id, e.height)
+            },
 
-    #[error("failed while querying Tx for client {0} on chain id: {1} with error: {2}")]
-    ClientEventQuery(ClientId, ChainId, String),
+        MissingSmallerTrustedHeight
+            {
+                chain_id: ChainId,
+                target_height: Height,
+            }
+            |e| {
+                format_args!("chain {} is missing trusted state smaller than target height {}",
+                    e.chain_id, e.target_height)
+            },
 
-    #[error("failed while finding client {0}: expected chain_id in client state: {1}; actual chain_id: {2}")]
-    ClientFind(ClientId, ChainId, ChainId),
+        MissingTrustedHeight
+            {
+                chain_id: ChainId,
+                target_height: Height,
+            }
+            |e| {
+                format_args!("chain {} is missing trusted state at target height {}",
+                    e.chain_id, e.target_height)
+            },
 
-    #[error("client {0} on chain id {1} is expired or frozen")]
-    ExpiredOrFrozen(ClientId, ChainId),
+        ClientRefresh
+            {
+                client_id: ClientId,
+                reason: String
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("error raised while trying to refresh client {0}: {1}",
+                    e.client_id, e.reason)
+            },
 
-    #[error("error raised while checking for misbehaviour evidence: {0}")]
-    Misbehaviour(String),
+        ClientQuery
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("failed while querying Tx for client {0} on chain id: {1}",
+                    e.client_id, e.chain_id)
+            },
 
-    #[error("cannot run misbehaviour: {0}")]
-    MisbehaviourExit(String),
+        ClientUpgrade
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+                description: String,
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("failed while trying to upgrade client id {0} for chain {1}: {2}: {3}",
+                    e.client_id, e.chain_id, e.description, e.source)
+            },
 
-    #[error("failed while trying to upgrade client id {0} with error: {1}")]
-    ClientUpgrade(ClientId, String),
+        ClientEventQuery
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+                consensus_height: Height
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("failed while querying Tx for client {0} on chain id {1} at consensus height {2}",
+                    e.client_id, e.chain_id, e.consensus_height)
+            },
+
+        UnexpectedEvent
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+                event: String,
+            }
+            |e| {
+                format_args!("failed while querying Tx for client {0} on chain id {1}: query Tx-es returned unexpected event: {2}",
+                    e.client_id, e.chain_id, e.event)
+            },
+
+        MismatchChainId
+            {
+                client_id: ClientId,
+                expected_chain_id: ChainId,
+                actual_chain_id: ChainId,
+            }
+            |e| {
+                format_args!("failed while finding client {0}: expected chain_id in client state: {1}; actual chain_id: {2}",
+                    e.client_id, e.expected_chain_id, e.actual_chain_id)
+            },
+
+        ExpiredOrFrozen
+            {
+                client_id: ClientId,
+                chain_id: ChainId,
+            }
+            |e| {
+                format_args!("client {0} on chain id {1} is expired or frozen",
+                    e.client_id, e.chain_id)
+            },
+
+        Misbehaviour
+            {
+                description: String,
+            }
+            [ RelayerError ]
+            |e| {
+                format_args!("error raised while checking for misbehaviour evidence: {0}", e.description)
+            },
+
+        MisbehaviourExit
+            { reason: String }
+            |e| {
+                format_args!("cannot run misbehaviour: {0}", e.reason)
+            },
+
+        SameChainId
+            {
+                chain_id: ChainId
+            }
+            |e| {
+                format_args!("the chain ID ({}) at the source and destination chains must be different", e.chain_id)
+            },
+
+        MissingClientIdFromEvent
+            { event: IbcEvent }
+            |e| {
+                format_args!("cannot extract client_id from result: {:?}",
+                    e.event)
+            },
+
+        ChainErrorEvent
+            {
+                chain_id: ChainId,
+                event: IbcEvent
+            }
+            |e| {
+                format_args!("Failed to update client on destination {} because of error event: {}",
+                    e.chain_id, e.event)
+            },
+    }
 }
 
 #[derive(Clone, Debug)]
-pub struct ForeignClient {
+pub struct ForeignClient<DstChain: ChainHandle, SrcChain: ChainHandle> {
     /// The identifier of this client. The host chain determines this id upon client creation,
     /// so we may be using the default value temporarily.
     pub id: ClientId,
 
     /// A handle to the chain hosting this client, i.e., destination chain.
-    pub dst_chain: Box<dyn ChainHandle>,
+    pub dst_chain: DstChain,
 
     /// A handle to the chain whose headers this client is verifying, aka the source chain.
-    pub src_chain: Box<dyn ChainHandle>,
+    pub src_chain: SrcChain,
 }
 
 /// Used in Output messages.
@@ -85,7 +232,9 @@ pub struct ForeignClient {
 /// where the first chain identifier is for the source
 /// chain, and the second chain identifier is the
 /// destination (which hosts the client) chain.
-impl fmt::Display for ForeignClient {
+impl<DstChain: ChainHandle, SrcChain: ChainHandle> fmt::Display
+    for ForeignClient<DstChain, SrcChain>
+{
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(
             f,
@@ -97,27 +246,23 @@ impl fmt::Display for ForeignClient {
     }
 }
 
-impl ForeignClient {
+impl<DstChain: ChainHandle, SrcChain: ChainHandle> ForeignClient<DstChain, SrcChain> {
     /// Creates a new foreign client on `dst_chain`. Blocks until the client is created, or
     /// an error occurs.
     /// Post-condition: `dst_chain` hosts an IBC client for `src_chain`.
     pub fn new(
-        dst_chain: Box<dyn ChainHandle>,
-        src_chain: Box<dyn ChainHandle>,
-    ) -> Result<ForeignClient, ForeignClientError> {
+        dst_chain: DstChain,
+        src_chain: SrcChain,
+    ) -> Result<ForeignClient<DstChain, SrcChain>, ForeignClientError> {
         // Sanity check
         if src_chain.id().eq(&dst_chain.id()) {
-            return Err(ForeignClientError::ClientCreate(format!(
-                "the source ({}) and destination ({}) chains must be different",
-                src_chain.id(),
-                dst_chain.id(),
-            )));
+            return Err(ForeignClientError::same_chain_id(src_chain.id()));
         }
 
         let mut client = ForeignClient {
             id: ClientId::default(),
-            dst_chain: dst_chain.clone(),
-            src_chain: src_chain.clone(),
+            dst_chain,
+            src_chain,
         };
 
         client.create()?;
@@ -127,9 +272,9 @@ impl ForeignClient {
 
     pub fn restore(
         id: ClientId,
-        dst_chain: Box<dyn ChainHandle>,
-        src_chain: Box<dyn ChainHandle>,
-    ) -> ForeignClient {
+        dst_chain: DstChain,
+        src_chain: SrcChain,
+    ) -> ForeignClient<DstChain, SrcChain> {
         ForeignClient {
             id,
             dst_chain,
@@ -143,16 +288,16 @@ impl ForeignClient {
     /// verifying) is consistent with `expected_target_chain`, and if so, return a new
     /// `ForeignClient` representing this client.
     pub fn find(
-        expected_target_chain: Box<dyn ChainHandle>,
-        host_chain: Box<dyn ChainHandle>,
+        expected_target_chain: SrcChain,
+        host_chain: DstChain,
         client_id: &ClientId,
-    ) -> Result<ForeignClient, ForeignClientError> {
+    ) -> Result<ForeignClient<DstChain, SrcChain>, ForeignClientError> {
         let height = Height::new(expected_target_chain.id().version(), 0);
 
         match host_chain.query_client_state(client_id, height) {
             Ok(cs) => {
                 if cs.chain_id() != expected_target_chain.id() {
-                    Err(ForeignClientError::ClientFind(
+                    Err(ForeignClientError::mismatch_chain_id(
                         client_id.clone(),
                         expected_target_chain.id(),
                         cs.chain_id(),
@@ -161,15 +306,15 @@ impl ForeignClient {
                     // TODO: Any additional checks?
                     Ok(ForeignClient::restore(
                         client_id.clone(),
-                        host_chain.clone(),
+                        host_chain,
                         expected_target_chain,
                     ))
                 }
             }
-            Err(e) => Err(ForeignClientError::ClientQuery(
+            Err(e) => Err(ForeignClientError::client_query(
                 client_id.clone(),
                 host_chain.id(),
-                format!("{}", e),
+                e,
             )),
         }
     }
@@ -177,13 +322,11 @@ impl ForeignClient {
     pub fn upgrade(&self) -> Result<Vec<IbcEvent>, ForeignClientError> {
         // Fetch the latest height of the source chain.
         let src_height = self.src_chain.query_latest_height().map_err(|e| {
-            ForeignClientError::ClientUpgrade(
+            ForeignClientError::client_upgrade(
                 self.id.clone(),
-                format!(
-                    "failed while querying src chain ({}) for latest height: {}",
-                    self.src_chain.id(),
-                    e
-                ),
+                self.src_chain.id(),
+                "failed while querying src chain for latest height".to_string(),
+                e,
             )
         })?;
 
@@ -196,13 +339,11 @@ impl ForeignClient {
             .src_chain
             .query_upgraded_client_state(src_height)
             .map_err(|e| {
-                ForeignClientError::ClientUpgrade(
+                ForeignClientError::client_upgrade(
                     self.id.clone(),
-                    format!(
-                        "failed while fetching from chain {} the upgraded client state: {}",
-                        self.src_chain.id(),
-                        e
-                    ),
+                    self.src_chain.id(),
+                    "failed while fetching from chain the upgraded client state".to_string(),
+                    e,
                 )
             })?;
 
@@ -211,9 +352,15 @@ impl ForeignClient {
         let (consensus_state, proof_upgrade_consensus_state) = self
             .src_chain
             .query_upgraded_consensus_state(src_height)
-            .map_err(|e| ForeignClientError::ClientUpgrade(self.id.clone(), format!(
-                "failed while fetching from chain {} the upgraded client consensus state: {}", self.src_chain.id(), e)))
-            ?;
+            .map_err(|e| {
+                ForeignClientError::client_upgrade(
+                    self.id.clone(),
+                    self.src_chain.id(),
+                    "failed while fetching from chain the upgraded client consensus state"
+                        .to_string(),
+                    e,
+                )
+            })?;
 
         debug!(
             "[{}]  upgraded client consensus state {:?}",
@@ -222,13 +369,11 @@ impl ForeignClient {
 
         // Get signer
         let signer = self.dst_chain.get_signer().map_err(|e| {
-            ForeignClientError::ClientUpgrade(
+            ForeignClientError::client_upgrade(
                 self.id.clone(),
-                format!(
-                    "failed while fetching the destination chain ({}) signer: {}",
-                    self.dst_chain.id(),
-                    e
-                ),
+                self.dst_chain.id(),
+                "failed while fetching the destination chain signer".to_string(),
+                e,
             )
         })?;
 
@@ -244,27 +389,28 @@ impl ForeignClient {
 
         msgs.push(msg_upgrade);
 
-        let res = self.dst_chain.send_msgs(msgs).map_err(|e| {
-            ForeignClientError::ClientUpgrade(
-                self.id.clone(),
-                format!(
-                    "failed while sending message to destination chain {} with err: {}",
+        let res = self
+            .dst_chain
+            .send_messages_and_wait_commit(msgs)
+            .map_err(|e| {
+                ForeignClientError::client_upgrade(
+                    self.id.clone(),
                     self.dst_chain.id(),
-                    e
-                ),
-            )
-        })?;
+                    "failed while sending message to destination chain".to_string(),
+                    e,
+                )
+            })?;
 
         Ok(res)
     }
 
     /// Returns a handle to the chain hosting this client.
-    pub fn dst_chain(&self) -> Box<dyn ChainHandle> {
+    pub fn dst_chain(&self) -> DstChain {
         self.dst_chain.clone()
     }
 
     /// Returns a handle to the chain whose headers this client is sourcing (the source chain).
-    pub fn src_chain(&self) -> Box<dyn ChainHandle> {
+    pub fn src_chain(&self) -> SrcChain {
         self.src_chain.clone()
     }
 
@@ -276,46 +422,53 @@ impl ForeignClient {
     pub fn build_create_client(&self) -> Result<MsgCreateAnyClient, ForeignClientError> {
         // Get signer
         let signer = self.dst_chain.get_signer().map_err(|e| {
-            ForeignClientError::ClientCreate(format!(
-                "failed while fetching the destination chain ({}) signer: {}",
+            ForeignClientError::client_create(
                 self.dst_chain.id(),
-                e
-            ))
+                "failed while fetching the destination chain signer".to_string(),
+                e,
+            )
         })?;
 
         // Build client create message with the data from source chain at latest height.
         let latest_height = self.src_chain.query_latest_height().map_err(|e| {
-            ForeignClientError::ClientCreate(format!(
-                "failed while querying src chain ({}) for latest height: {}",
-                self.src_chain.id(),
-                e
-            ))
+            ForeignClientError::client_create(
+                self.dst_chain.id(),
+                "failed while querying src chain ({}) for latest height: {}".to_string(),
+                e,
+            )
         })?;
 
         let client_state = self
             .src_chain
             .build_client_state(latest_height)
             .map_err(|e| {
-                ForeignClientError::ClientCreate(format!(
-                    "failed while building client state from src chain ({}) with error: {}",
+                ForeignClientError::client_create(
                     self.src_chain.id(),
-                    e
-                ))
+                    "failed while querying src chain for latest height".to_string(),
+                    e,
+                )
             })?
             .wrap_any();
 
-        let consensus_state = self.src_chain
-            .build_consensus_state(client_state.latest_height(),  latest_height, client_state.clone())
-            .map_err(|e| ForeignClientError::ClientCreate(format!("failed while building client consensus state from src chain ({}) with error: {}", self.src_chain.id(), e)))?
+        let consensus_state = self
+            .src_chain
+            .build_consensus_state(
+                client_state.latest_height(),
+                latest_height,
+                client_state.clone(),
+            )
+            .map_err(|e| {
+                ForeignClientError::client_create(
+                    self.src_chain.id(),
+                    "failed while building client consensus state from src chain".to_string(),
+                    e,
+                )
+            })?
             .wrap_any();
 
         //TODO Get acct_prefix
-        let msg = MsgCreateAnyClient::new(client_state, consensus_state, signer).map_err(|e| {
-            ForeignClientError::ClientCreate(format!(
-                "failed while building the create client message: {}",
-                e
-            ))
-        })?;
+        let msg = MsgCreateAnyClient::new(client_state, consensus_state, signer)
+            .map_err(ForeignClientError::client)?;
 
         Ok(msg)
     }
@@ -326,13 +479,13 @@ impl ForeignClient {
 
         let res = self
             .dst_chain
-            .send_msgs(vec![new_msg.to_any()])
+            .send_messages_and_wait_commit(vec![new_msg.to_any()])
             .map_err(|e| {
-                ForeignClientError::ClientCreate(format!(
-                    "failed sending message to dst chain ({}) with err: {}",
+                ForeignClientError::client_create(
                     self.dst_chain.id(),
-                    e
-                ))
+                    "failed sending message to dst chain ".to_string(),
+                    e,
+                )
             })?;
 
         assert!(!res.is_empty());
@@ -341,19 +494,14 @@ impl ForeignClient {
 
     /// Sends the client creation transaction & subsequently sets the id of this ForeignClient
     fn create(&mut self) -> Result<(), ForeignClientError> {
-        match self.build_create_client_and_send() {
-            Err(e) => {
-                error!("[{}]  failed CreateClient: {}", self, e);
-                return Err(ForeignClientError::ClientCreate(format!(
-                    "Create client failed ({:?})",
-                    e
-                )));
-            }
-            Ok(event) => {
-                self.id = extract_client_id(&event)?.clone();
-                info!("🍭 [{}]  => {:#?}\n", self, event);
-            }
-        }
+        let event = self.build_create_client_and_send().map_err(|e| {
+            error!("[{}]  failed CreateClient: {}", self, e);
+            e
+        })?;
+
+        self.id = extract_client_id(&event)?.clone();
+        info!("🍭 [{}]  => {:#?}\n", self, event);
+
         Ok(())
     }
 
@@ -362,10 +510,11 @@ impl ForeignClient {
             .dst_chain
             .query_client_state(self.id(), Height::zero())
             .map_err(|e| {
-                ForeignClientError::ClientUpdate(format!(
-                    "failed querying client state on dst chain {} with error: {}",
-                    self.id, e
-                ))
+                ForeignClientError::client_refresh(
+                    self.id().clone(),
+                    "failed querying client state on dst chain".to_string(),
+                    e,
+                )
             })?;
 
         let last_update_time = self
@@ -376,7 +525,7 @@ impl ForeignClient {
         let elapsed = Timestamp::now().duration_since(&last_update_time);
 
         if client_state.is_frozen() || client_state.expired(elapsed.unwrap_or_default()) {
-            return Err(ForeignClientError::ExpiredOrFrozen(
+            return Err(ForeignClientError::expired_or_frozen(
                 self.id().clone(),
                 self.dst_chain.id(),
             ));
@@ -413,10 +562,11 @@ impl ForeignClient {
     ) -> Result<Vec<Any>, ForeignClientError> {
         // Wait for source chain to reach `target_height`
         while self.src_chain().query_latest_height().map_err(|e| {
-            ForeignClientError::ClientUpdate(format!(
-                "failed fetching src chain latest height with error: {}",
-                e
-            ))
+            ForeignClientError::client_create(
+                self.src_chain.id(),
+                "failed fetching src chain latest height with error".to_string(),
+                e,
+            )
         })? < target_height
         {
             thread::sleep(Duration::from_millis(100))
@@ -427,10 +577,11 @@ impl ForeignClient {
             .dst_chain()
             .query_client_state(&self.id, Height::default())
             .map_err(|e| {
-                ForeignClientError::ClientUpdate(format!(
-                    "failed querying client state on dst chain {} with error: {}",
-                    self.id, e
-                ))
+                ForeignClientError::client_create(
+                    self.dst_chain.id(),
+                    "failed querying client state on dst chain".to_string(),
+                    e,
+                )
             })?;
 
         // If not specified, set trusted state to the highest height smaller than target height.
@@ -442,22 +593,18 @@ impl ForeignClient {
                 .into_iter()
                 .find(|h| h < &target_height)
                 .ok_or_else(|| {
-                    ForeignClientError::ClientUpdate(format!(
-                        "chain {} is missing trusted state smaller than target height {}",
+                    ForeignClientError::missing_smaller_trusted_height(
                         self.dst_chain().id(),
-                        target_height
-                    ))
+                        target_height,
+                    )
+                    // ))
                 })?
         } else {
             cs_heights
                 .into_iter()
                 .find(|h| h == &trusted_height)
                 .ok_or_else(|| {
-                    ForeignClientError::ClientUpdate(format!(
-                        "chain {} is missing trusted state at height {}",
-                        self.dst_chain().id(),
-                        trusted_height
-                    ))
+                    ForeignClientError::missing_trusted_height(self.dst_chain().id(), target_height)
                 })?
         };
 
@@ -473,18 +620,19 @@ impl ForeignClient {
             .src_chain()
             .build_header(trusted_height, target_height, client_state)
             .map_err(|e| {
-                ForeignClientError::ClientUpdate(format!(
-                    "failed building header with error: {}",
-                    e
-                ))
+                ForeignClientError::client_update(
+                    self.src_chain.id(),
+                    "failed building header with error".to_string(),
+                    e,
+                )
             })?;
 
         let signer = self.dst_chain().get_signer().map_err(|e| {
-            ForeignClientError::ClientUpdate(format!(
-                "failed getting signer for dst chain ({}) with error: {}",
+            ForeignClientError::client_update(
                 self.dst_chain.id(),
-                e
-            ))
+                "failed getting signer for dst chain".to_string(),
+                e,
+            )
         })?;
 
         let mut msgs = vec![];
@@ -536,11 +684,11 @@ impl ForeignClient {
     ) -> Result<Vec<IbcEvent>, ForeignClientError> {
         let h = if height == Height::zero() {
             self.src_chain.query_latest_height().map_err(|e| {
-                ForeignClientError::ClientUpdate(format!(
-                    "failed while querying src chain ({}) for latest height: {}",
+                ForeignClientError::client_update(
                     self.src_chain.id(),
-                    e
-                ))
+                    "failed while querying src chain ({}) for latest height".to_string(),
+                    e,
+                )
             })?
         } else {
             height
@@ -548,30 +696,30 @@ impl ForeignClient {
 
         let new_msgs = self.build_update_client_with_trusted(h, trusted_height)?;
         if new_msgs.is_empty() {
-            return Err(ForeignClientError::ClientUpdate(format!(
-                "Client {} is already up-to-date with chain {}@{}",
-                self.id,
+            return Err(ForeignClientError::client_already_up_to_date(
+                self.id.clone(),
                 self.src_chain.id(),
-                h
-            )));
+                h,
+            ));
         }
 
-        let events = self.dst_chain().send_msgs(new_msgs).map_err(|e| {
-            ForeignClientError::ClientUpdate(format!(
-                "failed sending message to dst chain ({}) with err: {}",
-                self.dst_chain.id(),
-                e
-            ))
-        })?;
+        let events = self
+            .dst_chain()
+            .send_messages_and_wait_commit(new_msgs)
+            .map_err(|e| {
+                ForeignClientError::client_update(
+                    self.dst_chain.id(),
+                    "failed sending message to dst chain".to_string(),
+                    e,
+                )
+            })?;
 
         Ok(events)
     }
 
     /// Attempts to update a client using header from the latest height of its source chain.
     pub fn update(&self) -> Result<(), ForeignClientError> {
-        let res = self.build_latest_update_client_and_send().map_err(|e| {
-            ForeignClientError::ClientUpdate(format!("build_create_client_and_send {:?}", e))
-        })?;
+        let res = self.build_latest_update_client_and_send()?;
 
         debug!("[{}] client updated with return message {:?}\n", self, res);
 
@@ -600,10 +748,11 @@ impl ForeignClient {
                 .dst_chain
                 .query_txs(QueryTxRequest::Client(request.clone()))
                 .map_err(|e| {
-                    ForeignClientError::ClientEventQuery(
+                    ForeignClientError::client_event_query(
                         self.id().clone(),
                         self.dst_chain.id(),
-                        format!("update event for {}: {}", consensus_height, e),
+                        consensus_height,
+                        e,
                     )
                 });
             match result {
@@ -633,10 +782,10 @@ impl ForeignClient {
         // Regardless, just take the event from the first update.
         let event = events[0].clone();
         let update = downcast!(event.clone() => IbcEvent::UpdateClient).ok_or_else(|| {
-            ForeignClientError::ClientEventQuery(
+            ForeignClientError::unexpected_event(
                 self.id().clone(),
                 self.dst_chain.id(),
-                format!("query Tx-es returned unexpected event {}", event.to_json()),
+                event.to_json(),
             )
         })?;
         Ok(Some(update))
@@ -653,11 +802,7 @@ impl ForeignClient {
                 pagination: ibc_proto::cosmos::base::query::pagination::all(),
             })
             .map_err(|e| {
-                ForeignClientError::ClientQuery(
-                    self.id().clone(),
-                    self.src_chain.id(),
-                    format!("{}", e),
-                )
+                ForeignClientError::client_query(self.id().clone(), self.src_chain.id(), e)
             })?;
         consensus_states.sort_by_key(|a| std::cmp::Reverse(a.height));
         Ok(consensus_states)
@@ -669,14 +814,7 @@ impl ForeignClient {
             .dst_chain
             .query_consensus_state(self.id.clone(), height, Height::zero())
             .map_err(|e| {
-                ForeignClientError::ClientQuery(
-                    self.id.clone(),
-                    self.dst_chain.id(),
-                    format!(
-                        "failed querying consensus state at height {} with error {}",
-                        height, e
-                    ),
-                )
+                ForeignClientError::client_query(self.id.clone(), self.dst_chain.id(), e)
             })?;
 
         Ok(res)
@@ -739,10 +877,10 @@ impl ForeignClient {
             .dst_chain()
             .query_client_state(&self.id, Height::zero())
             .map_err(|e| {
-                ForeignClientError::Misbehaviour(format!(
-                    "failed querying client state on dst chain {} with error: {}",
-                    self.id, e
-                ))
+                ForeignClientError::misbehaviour(
+                    format!("failed querying client state on dst chain {}", self.id),
+                    e,
+                )
             })?;
 
         // Get the list of consensus state heights in descending order.
@@ -805,7 +943,7 @@ impl ForeignClient {
             // No header in events, cannot run misbehavior.
             // May happen on chains running older SDKs (e.g., Akash)
             if update_event.header.is_none() {
-                return Err(ForeignClientError::MisbehaviourExit(
+                return Err(ForeignClientError::misbehaviour_exit(
                     "no header in update client events".to_string(),
                 ));
             }
@@ -817,12 +955,14 @@ impl ForeignClient {
                 .src_chain
                 .check_misbehaviour(update_event.clone(), client_state.clone())
                 .map_err(|e| {
-                    ForeignClientError::Misbehaviour(format!(
-                        "failed to check misbehaviour for {} at consensus height {}: {}",
-                        update_event.client_id(),
-                        update_event.consensus_height(),
-                        e
-                    ))
+                    ForeignClientError::misbehaviour(
+                        format!(
+                            "failed to check misbehaviour for {} at consensus height {}",
+                            update_event.client_id(),
+                            update_event.consensus_height(),
+                        ),
+                        e,
+                    )
                 })?;
 
             if misbehavior.is_some() {
@@ -855,11 +995,13 @@ impl ForeignClient {
         evidence: MisbehaviourEvidence,
     ) -> Result<Vec<IbcEvent>, ForeignClientError> {
         let signer = self.dst_chain().get_signer().map_err(|e| {
-            ForeignClientError::Misbehaviour(format!(
-                "failed getting signer for destination chain ({}), error: {}",
-                self.dst_chain.id(),
-                e
-            ))
+            ForeignClientError::misbehaviour(
+                format!(
+                    "failed getting signer for destination chain ({})",
+                    self.dst_chain.id()
+                ),
+                e,
+            )
         })?;
 
         let mut msgs = vec![];
@@ -884,13 +1026,18 @@ impl ForeignClient {
             .to_any(),
         );
 
-        let events = self.dst_chain().send_msgs(msgs).map_err(|e| {
-            ForeignClientError::Misbehaviour(format!(
-                "failed sending evidence to destination chain ({}), error: {}",
-                self.dst_chain.id(),
-                e
-            ))
-        })?;
+        let events = self
+            .dst_chain()
+            .send_messages_and_wait_commit(msgs)
+            .map_err(|e| {
+                ForeignClientError::misbehaviour(
+                    format!(
+                        "failed sending evidence to destination chain ({})",
+                        self.dst_chain.id(),
+                    ),
+                    e,
+                )
+            })?;
 
         Ok(events)
     }
@@ -917,7 +1064,7 @@ impl ForeignClient {
         // Even if some states may have failed to verify, e.g. if they were expired, just
         // warn the user and continue.
         match result {
-            Err(ForeignClientError::MisbehaviourExit(s)) => {
+            Err(ForeignClientError(ForeignClientErrorDetail::MisbehaviourExit(s), _)) => {
                 warn!(
                     "[{}] misbehaviour checking is being disabled: {:?}",
                     self, s
@@ -935,14 +1082,23 @@ impl ForeignClient {
                     MisbehaviourResults::ValidClient
                 }
             }
-            Err(e) => {
-                if update_event.is_some() {
+            Err(e) => match e.detail() {
+                ForeignClientErrorDetail::MisbehaviourExit(s) => {
+                    error!(
+                        "[{}] misbehaviour checking is being disabled: {:?}",
+                        self, s
+                    );
                     MisbehaviourResults::CannotExecute
-                } else {
-                    warn!("[{}] misbehaviour checking result {:?}", self, e);
-                    MisbehaviourResults::ValidClient
                 }
-            }
+                _ => {
+                    if update_event.is_some() {
+                        MisbehaviourResults::CannotExecute
+                    } else {
+                        warn!("[{}] misbehaviour checking result {:?}", self, e);
+                        MisbehaviourResults::ValidClient
+                    }
+                }
+            },
         }
     }
 }
@@ -959,10 +1115,9 @@ pub fn extract_client_id(event: &IbcEvent) -> Result<&ClientId, ForeignClientErr
     match event {
         IbcEvent::CreateClient(ev) => Ok(ev.client_id()),
         IbcEvent::UpdateClient(ev) => Ok(ev.client_id()),
-        other => Err(ForeignClientError::ClientCreate(format!(
-            "cannot extract client_id from result: {:?}",
-            other
-        ))),
+        _ => Err(ForeignClientError::missing_client_id_from_event(
+            event.clone(),
+        )),
     }
 }
 
@@ -981,6 +1136,7 @@ mod test {
     use ibc::ics24_host::identifier::ClientId;
     use ibc::Height;
 
+    use crate::chain::handle::{ChainHandle, ProdChainHandle};
     use crate::chain::mock::test_utils::get_basic_chain_config;
     use crate::chain::mock::MockChain;
     use crate::chain::runtime::ChainRuntime;
@@ -993,13 +1149,13 @@ mod test {
         let b_cfg = get_basic_chain_config("chain_b");
 
         let rt = Arc::new(TokioRuntime::new().unwrap());
-        let a_chain = ChainRuntime::<MockChain>::spawn(a_cfg, rt.clone()).unwrap();
-        let b_chain = ChainRuntime::<MockChain>::spawn(b_cfg, rt).unwrap();
+        let a_chain =
+            ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(a_cfg, rt.clone()).unwrap();
+        let b_chain = ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(b_cfg, rt).unwrap();
         let a_client =
             ForeignClient::restore(ClientId::default(), a_chain.clone(), b_chain.clone());
 
-        let b_client =
-            ForeignClient::restore(ClientId::default(), b_chain.clone(), a_chain.clone());
+        let b_client = ForeignClient::restore(ClientId::default(), b_chain, a_chain);
 
         // Create the client on chain a
         let res = a_client.build_create_client_and_send();
@@ -1031,8 +1187,9 @@ mod test {
         let num_iterations = 3;
 
         let rt = Arc::new(TokioRuntime::new().unwrap());
-        let a_chain = ChainRuntime::<MockChain>::spawn(a_cfg, rt.clone()).unwrap();
-        let b_chain = ChainRuntime::<MockChain>::spawn(b_cfg, rt).unwrap();
+        let a_chain =
+            ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(a_cfg, rt.clone()).unwrap();
+        let b_chain = ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(b_cfg, rt).unwrap();
         let mut a_client = ForeignClient::restore(a_client_id, a_chain.clone(), b_chain.clone());
 
         let mut b_client =
@@ -1046,7 +1203,7 @@ mod test {
         );
 
         // Remember b's height.
-        let b_height_start = b_chain.clone().query_latest_height().unwrap();
+        let b_height_start = b_chain.query_latest_height().unwrap();
 
         // Create a client on chain a
         let res = a_client.create();
@@ -1137,8 +1294,9 @@ mod test {
         let b_cfg = get_basic_chain_config("chain_b");
 
         let rt = Arc::new(TokioRuntime::new().unwrap());
-        let a_chain = ChainRuntime::<MockChain>::spawn(a_cfg, rt.clone()).unwrap();
-        let b_chain = ChainRuntime::<MockChain>::spawn(b_cfg, rt).unwrap();
+        let a_chain =
+            ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(a_cfg, rt.clone()).unwrap();
+        let b_chain = ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(b_cfg, rt).unwrap();
 
         // Instantiate the foreign clients on the two chains.
         let res_client_on_a = ForeignClient::new(a_chain.clone(), b_chain.clone());
@@ -1185,8 +1343,9 @@ mod test {
         let mut _b_client_id = ClientId::from_str("client_on_b_fora").unwrap();
 
         let rt = Arc::new(TokioRuntime::new().unwrap());
-        let a_chain = ChainRuntime::<MockChain>::spawn(a_cfg, rt.clone()).unwrap();
-        let b_chain = ChainRuntime::<MockChain>::spawn(b_cfg, rt).unwrap();
+        let a_chain =
+            ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(a_cfg, rt.clone()).unwrap();
+        let b_chain = ChainRuntime::<MockChain>::spawn::<ProdChainHandle>(b_cfg, rt).unwrap();
 
         // Instantiate the foreign clients on the two chains.
         let client_on_a_res = ForeignClient::new(a_chain.clone(), b_chain.clone());
