@@ -1,4 +1,4 @@
-use std::convert::TryFrom;
+use std::convert::{TryFrom, TryInto};
 
 use itertools::Itertools;
 
@@ -30,12 +30,7 @@ use ibc::{
 };
 use tracing::trace;
 
-use crate::error::Kind;
-use crate::{
-    chain::CosmosSdkChain,
-    config::ChainConfig,
-    error::{self, Error},
-};
+use crate::{chain::CosmosSdkChain, config::ChainConfig, error::Error};
 
 use super::Verified;
 
@@ -65,8 +60,8 @@ impl super::LightClient<CosmosSdkChain> for LightClient {
     ) -> Result<Verified<LightBlock>, Error> {
         trace!(%trusted, %target, "light client verification");
 
-        let target_height = TMHeight::try_from(target.revision_height)
-            .map_err(|e| error::Kind::InvalidHeight.context(e))?;
+        let target_height =
+            TMHeight::try_from(target.revision_height).map_err(Error::invalid_height)?;
 
         let client = self.prepare_client(client_state)?;
         let mut state = self.prepare_state(trusted)?;
@@ -74,7 +69,7 @@ impl super::LightClient<CosmosSdkChain> for LightClient {
         // Verify the target header
         let target = client
             .verify_to_target(target_height, &mut state)
-            .map_err(|e| error::Kind::LightClient(self.chain_id.to_string()).context(e))?;
+            .map_err(|e| Error::light_client(self.chain_id.to_string(), e))?;
 
         // Collect the verification trace for the target block
         let target_trace = state.get_trace(target.height());
@@ -93,8 +88,7 @@ impl super::LightClient<CosmosSdkChain> for LightClient {
     fn fetch(&mut self, height: ibc::Height) -> Result<LightBlock, Error> {
         trace!(%height, "fetching header");
 
-        let height = TMHeight::try_from(height.revision_height)
-            .map_err(|e| error::Kind::InvalidHeight.context(e))?;
+        let height = TMHeight::try_from(height.revision_height).map_err(Error::invalid_height)?;
 
         self.fetch_light_block(AtHeight::At(height))
     }
@@ -112,14 +106,14 @@ impl super::LightClient<CosmosSdkChain> for LightClient {
         crate::time!("light client check_misbehaviour");
 
         let update_header = update.header.clone().ok_or_else(|| {
-            Kind::Misbehaviour(format!(
+            Error::misbehaviour(format!(
                 "missing header in update client event {}",
                 self.chain_id
             ))
         })?;
 
         let update_header = downcast!(update_header => AnyHeader::Tendermint).ok_or_else(|| {
-            Kind::Misbehaviour(format!(
+            Error::misbehaviour(format!(
                 "header type incompatible for chain {}",
                 self.chain_id
             ))
@@ -172,7 +166,7 @@ impl super::LightClient<CosmosSdkChain> for LightClient {
 impl LightClient {
     pub fn from_config(config: &ChainConfig, peer_id: PeerId) -> Result<Self, Error> {
         let rpc_client = rpc::HttpClient::new(config.rpc_addr.clone())
-            .map_err(|e| error::Kind::LightClient(config.rpc_addr.to_string()).context(e))?;
+            .map_err(|e| Error::rpc(config.rpc_addr.clone(), e))?;
 
         let io = components::io::ProdIo::new(peer_id, rpc_client, Some(config.rpc_timeout));
 
@@ -191,14 +185,14 @@ impl LightClient {
 
         let client_state =
             downcast!(client_state => AnyClientState::Tendermint).ok_or_else(|| {
-                error::Kind::ClientTypeMismatch {
-                    expected: ClientType::Tendermint,
-                    got: client_state.client_type(),
-                }
+                Error::client_type_mismatch(ClientType::Tendermint, client_state.client_type())
             })?;
 
         let params = TmOptions {
-            trust_threshold: client_state.trust_level,
+            trust_threshold: client_state
+                .trust_level
+                .try_into()
+                .map_err(Error::light_client_state)?,
             trusting_period: client_state.trusting_period,
             clock_drift: client_state.max_clock_drift,
         };
@@ -215,8 +209,8 @@ impl LightClient {
     }
 
     fn prepare_state(&self, trusted: ibc::Height) -> Result<LightClientState, Error> {
-        let trusted_height = TMHeight::try_from(trusted.revision_height)
-            .map_err(|e| error::Kind::InvalidHeight.context(e))?;
+        let trusted_height =
+            TMHeight::try_from(trusted.revision_height).map_err(Error::invalid_height)?;
 
         let trusted_block = self.fetch_light_block(AtHeight::At(trusted_height))?;
 
@@ -229,11 +223,9 @@ impl LightClient {
     fn fetch_light_block(&self, height: AtHeight) -> Result<LightBlock, Error> {
         use tendermint_light_client::components::io::Io;
 
-        self.io.fetch_light_block(height).map_err(|e| {
-            error::Kind::LightClient(self.chain_id.to_string())
-                .context(e)
-                .into()
-        })
+        self.io
+            .fetch_light_block(height)
+            .map_err(|e| Error::light_client_io(self.chain_id.to_string(), e))
     }
 
     fn adjust_headers(
