@@ -1,10 +1,10 @@
 use std::{
     collections::HashMap,
+    ops::Deref,
     sync::{Arc, RwLock},
     time::Duration,
 };
 
-use anomaly::BoxError;
 use crossbeam_channel::{Receiver, Sender};
 use itertools::Itertools;
 use tracing::{debug, error, info, trace, warn};
@@ -19,20 +19,20 @@ use crate::{
     chain::handle::ChainHandle,
     config::{ChainConfig, Config},
     event,
-    event::monitor::{EventBatch, UnwrapOrClone},
+    event::monitor::{Error as EventError, ErrorDetail as EventErrorDetail, EventBatch},
     object::Object,
     registry::Registry,
+    rest,
     telemetry::Telemetry,
     util::try_recv_multiple,
     worker::{WorkerMap, WorkerMsg},
 };
 
 pub mod client_state_filter;
-mod error;
-
 use client_state_filter::{FilterPolicy, Permission};
 
-pub use error::Error;
+pub mod error;
+pub use error::{Error, ErrorDetail};
 
 pub mod dump_state;
 use dump_state::SupervisorState;
@@ -47,29 +47,33 @@ use self::spawn::SpawnMode;
 
 type ArcBatch = Arc<event::monitor::Result<EventBatch>>;
 type Subscription = Receiver<ArcBatch>;
-type BoxHandle = Box<dyn ChainHandle>;
 
 pub type RwArc<T> = Arc<RwLock<T>>;
 
 /// The supervisor listens for events on multiple pairs of chains,
 /// and dispatches the events it receives to the appropriate
 /// worker, based on the [`Object`] associated with each event.
-pub struct Supervisor {
+pub struct Supervisor<Chain: ChainHandle> {
     config: RwArc<Config>,
-    registry: Registry,
+    registry: Registry<Chain>,
     workers: WorkerMap,
 
     cmd_rx: Receiver<SupervisorCmd>,
     worker_msg_rx: Receiver<WorkerMsg>,
+    rest_rx: Option<rest::Receiver>,
     client_state_filter: FilterPolicy,
 
     #[allow(dead_code)]
     telemetry: Telemetry,
 }
 
-impl Supervisor {
+impl<Chain: ChainHandle + 'static> Supervisor<Chain> {
     /// Create a [`Supervisor`] which will listen for events on all the chains in the [`Config`].
-    pub fn new(config: RwArc<Config>, telemetry: Telemetry) -> (Self, Sender<SupervisorCmd>) {
+    pub fn new(
+        config: RwArc<Config>,
+        rest_rx: Option<rest::Receiver>,
+        telemetry: Telemetry,
+    ) -> (Self, Sender<SupervisorCmd>) {
         let registry = Registry::new(config.clone());
         let (worker_msg_tx, worker_msg_rx) = crossbeam_channel::unbounded();
         let workers = WorkerMap::new(worker_msg_tx, telemetry.clone());
@@ -83,6 +87,7 @@ impl Supervisor {
             workers,
             cmd_rx,
             worker_msg_rx,
+            rest_rx,
             client_state_filter,
             telemetry,
         };
@@ -177,10 +182,10 @@ impl Supervisor {
     /// and maps each [`IbcEvent`] to their corresponding [`Object`].
     pub fn collect_events(
         &self,
-        src_chain: &dyn ChainHandle,
-        batch: EventBatch,
+        src_chain: &impl ChainHandle,
+        batch: &EventBatch,
     ) -> CollectedEvents {
-        let mut collected = CollectedEvents::new(batch.height, batch.chain_id);
+        let mut collected = CollectedEvents::new(batch.height, batch.chain_id.clone());
 
         let handshake_enabled = self
             .config
@@ -188,16 +193,20 @@ impl Supervisor {
             .expect("poisoned lock")
             .handshake_enabled();
 
-        for event in batch.events {
+        for event in &batch.events {
             match event {
                 IbcEvent::NewBlock(_) => {
-                    collected.new_block = Some(event);
+                    collected.new_block = Some(event.clone());
                 }
                 IbcEvent::UpdateClient(ref update) => {
                     if let Ok(object) = Object::for_update_client(update, src_chain) {
                         // Collect update client events only if the worker exists
                         if self.workers.contains(&object) {
-                            collected.per_object.entry(object).or_default().push(event);
+                            collected
+                                .per_object
+                                .entry(object)
+                                .or_default()
+                                .push(event.clone());
                         }
                     }
                 }
@@ -213,7 +222,11 @@ impl Supervisor {
                         .map(|attr| Object::connection_from_conn_open_events(attr, src_chain));
 
                     if let Some(Ok(object)) = object {
-                        collected.per_object.entry(object).or_default().push(event);
+                        collected
+                            .per_object
+                            .entry(object)
+                            .or_default()
+                            .push(event.clone());
                     }
                 }
                 IbcEvent::OpenInitChannel(..) | IbcEvent::OpenTryChannel(..) => {
@@ -226,7 +239,11 @@ impl Supervisor {
                         .map(|attr| Object::channel_from_chan_open_events(attr, src_chain));
 
                     if let Some(Ok(object)) = object {
-                        collected.per_object.entry(object).or_default().push(event);
+                        collected
+                            .per_object
+                            .entry(object)
+                            .or_default()
+                            .push(event.clone());
                     }
                 }
                 IbcEvent::OpenAckChannel(ref open_ack) => {
@@ -260,7 +277,7 @@ impl Supervisor {
                                 .per_object
                                 .entry(channel_object)
                                 .or_default()
-                                .push(event);
+                                .push(event.clone());
                         }
                     }
                 }
@@ -287,22 +304,38 @@ impl Supervisor {
                 }
                 IbcEvent::SendPacket(ref packet) => {
                     if let Ok(object) = Object::for_send_packet(packet, src_chain) {
-                        collected.per_object.entry(object).or_default().push(event);
+                        collected
+                            .per_object
+                            .entry(object)
+                            .or_default()
+                            .push(event.clone());
                     }
                 }
                 IbcEvent::TimeoutPacket(ref packet) => {
                     if let Ok(object) = Object::for_timeout_packet(packet, src_chain) {
-                        collected.per_object.entry(object).or_default().push(event);
+                        collected
+                            .per_object
+                            .entry(object)
+                            .or_default()
+                            .push(event.clone());
                     }
                 }
                 IbcEvent::WriteAcknowledgement(ref packet) => {
                     if let Ok(object) = Object::for_write_ack(packet, src_chain) {
-                        collected.per_object.entry(object).or_default().push(event);
+                        collected
+                            .per_object
+                            .entry(object)
+                            .or_default()
+                            .push(event.clone());
                     }
                 }
                 IbcEvent::CloseInitChannel(ref packet) => {
                     if let Ok(object) = Object::for_close_init_channel(packet, src_chain) {
-                        collected.per_object.entry(object).or_default().push(event);
+                        collected
+                            .per_object
+                            .entry(object)
+                            .or_default()
+                            .push(event.clone());
                     }
                 }
                 _ => (),
@@ -313,7 +346,7 @@ impl Supervisor {
     }
 
     /// Create a new `SpawnContext` for spawning workers.
-    fn spawn_context(&mut self, mode: SpawnMode) -> SpawnContext<'_> {
+    fn spawn_context(&mut self, mode: SpawnMode) -> SpawnContext<'_, Chain> {
         SpawnContext::new(
             &self.config,
             &mut self.registry,
@@ -330,7 +363,7 @@ impl Supervisor {
     }
 
     /// Run the supervisor event loop.
-    pub fn run(mut self) -> Result<(), BoxError> {
+    pub fn run(mut self) -> Result<(), Error> {
         self.spawn_workers(SpawnMode::Startup);
 
         let mut subscriptions = self.init_subscriptions()?;
@@ -352,18 +385,21 @@ impl Supervisor {
                         Ok(subs) => {
                             subscriptions = subs;
                         }
-                        Err(Error::NoChainsAvailable) => (),
-                        Err(e) => return Err(e.into()),
+                        Err(Error(ErrorDetail::NoChainsAvailable(_), _)) => (),
+                        Err(e) => return Err(e),
                     }
                 }
             }
 
+            // Process incoming requests from the REST server
+            self.handle_rest_requests();
+
             std::thread::sleep(Duration::from_millis(50));
         }
     }
 
     /// Subscribe to the events emitted by the chains the supervisor is connected to.
-    fn init_subscriptions(&mut self) -> Result<Vec<(BoxHandle, Subscription)>, Error> {
+    fn init_subscriptions(&mut self) -> Result<Vec<(Chain, Subscription)>, Error> {
         let chains = &self.config.read().expect("poisoned lock").chains;
 
         let mut subscriptions = Vec::with_capacity(chains.len());
@@ -392,7 +428,7 @@ impl Supervisor {
         // At least one chain runtime should be available, otherwise the supervisor
         // cannot do anything and will hang indefinitely.
         if self.registry.size() == 0 {
-            return Err(Error::NoChainsAvailable);
+            return Err(Error::no_chains_available());
         }
 
         Ok(subscriptions)
@@ -412,13 +448,19 @@ impl Supervisor {
     /// Dump the state of the supervisor into a [`SupervisorState`] value,
     /// and send it back through the given channel.
     fn dump_state(&self, reply_to: Sender<SupervisorState>) -> CmdEffect {
-        let chains = self.registry.chains().map(|c| c.id()).collect_vec();
-        let state = SupervisorState::new(chains, self.workers.objects());
+        let state = self.state();
         let _ = reply_to.try_send(state);
 
         CmdEffect::Nothing
     }
 
+    /// Returns a representation of the supervisor's internal state
+    /// as a [`SupervisorState`].
+    fn state(&self) -> SupervisorState {
+        let chains = self.registry.chains().map(|c| c.id()).collect_vec();
+        SupervisorState::new(chains, self.workers.objects())
+    }
+
     /// Apply the given configuration update.
     ///
     /// Returns an [`CmdEffect`] which instructs the caller as to
@@ -483,7 +525,7 @@ impl Supervisor {
     /// If the removal had any effect, returns [`CmdEffect::ConfigChanged`] as
     /// subscriptions need to be reset to take into account the newly added chain.
     fn remove_chain(&mut self, id: &ChainId) -> CmdEffect {
-        if !self.config.read().expect("poisoned lock").has_chain(&id) {
+        if !self.config.read().expect("poisoned lock").has_chain(id) {
             info!(chain.id=%id, "skipping removal of non-existing chain");
             return CmdEffect::Nothing;
         }
@@ -498,10 +540,10 @@ impl Supervisor {
 
         debug!(chain.id=%id, "shutting down workers");
         let mut ctx = self.spawn_context(SpawnMode::Reload);
-        ctx.shutdown_workers_for_chain(&id);
+        ctx.shutdown_workers_for_chain(id);
 
         debug!(chain.id=%id, "shutting down chain runtime");
-        self.registry.shutdown(&id);
+        self.registry.shutdown(id);
 
         CmdEffect::ConfigChanged
     }
@@ -529,33 +571,61 @@ impl Supervisor {
         }
     }
 
+    fn handle_rest_requests(&self) {
+        if let Some(rest_rx) = &self.rest_rx {
+            let config = self.config.read().expect("poisoned lock");
+            if let Some(cmd) = rest::process_incoming_requests(&config, rest_rx) {
+                self.handle_rest_cmd(cmd);
+            }
+        }
+    }
+
+    fn handle_rest_cmd(&self, m: rest::Command) {
+        match m {
+            rest::Command::DumpState(reply) => {
+                let state = self.state();
+                reply.send(Ok(state)).unwrap_or_else(|e| {
+                    error!("[rest/supervisor] error replying to a REST request {}", e)
+                });
+            }
+        }
+    }
+
     /// Process the given batch if it does not contain any errors,
     /// output the errors on the console otherwise.
-    fn handle_batch(&mut self, chain: Box<dyn ChainHandle>, batch: ArcBatch) {
+    fn handle_batch(&mut self, chain: Chain, batch: ArcBatch) {
         let chain_id = chain.id();
 
-        let result = batch
-            .unwrap_or_clone()
-            .map_err(Into::into)
-            .and_then(|batch| self.process_batch(chain, batch));
+        match batch.deref() {
+            Ok(batch) => {
+                let _ = self
+                    .process_batch(chain, batch)
+                    .map_err(|e| error!("[{}] error during batch processing: {}", chain_id, e));
+            }
+            Err(EventError(EventErrorDetail::SubscriptionCancelled(_), _)) => {
+                warn!(chain.id = %chain_id, "event subscription was cancelled, clearing pending packets");
 
-        if let Err(e) = result {
-            error!("[{}] error during batch processing: {}", chain_id, e);
+                let _ = self.clear_pending_packets(&chain_id).map_err(|e| {
+                    error!(
+                        "[{}] error during clearing pending packets: {}",
+                        chain_id, e
+                    )
+                });
+            }
+            Err(e) => {
+                error!("[{}] error in receiving event batch: {}", chain_id, e)
+            }
         }
     }
 
     /// Process a batch of events received from a chain.
-    fn process_batch(
-        &mut self,
-        src_chain: Box<dyn ChainHandle>,
-        batch: EventBatch,
-    ) -> Result<(), BoxError> {
+    fn process_batch(&mut self, src_chain: Chain, batch: &EventBatch) -> Result<(), Error> {
         assert_eq!(src_chain.id(), batch.chain_id);
 
         let height = batch.height;
         let chain_id = batch.chain_id.clone();
 
-        let mut collected = self.collect_events(src_chain.clone().as_ref(), batch);
+        let mut collected = self.collect_events(&src_chain, batch);
 
         for (object, events) in collected.per_object.drain() {
             if !self.relay_on_object(&src_chain.id(), &object) {
@@ -572,26 +642,45 @@ impl Supervisor {
                 continue;
             }
 
-            let src = self.registry.get_or_spawn(object.src_chain_id())?;
-            let dst = self.registry.get_or_spawn(object.dst_chain_id())?;
+            let src = self
+                .registry
+                .get_or_spawn(object.src_chain_id())
+                .map_err(Error::spawn)?;
+
+            let dst = self
+                .registry
+                .get_or_spawn(object.dst_chain_id())
+                .map_err(Error::spawn)?;
 
             let worker = {
                 let config = self.config.read().expect("poisoned lock");
                 self.workers.get_or_spawn(object, src, dst, &config)
             };
 
-            worker.send_events(height, events, chain_id.clone())?
+            worker
+                .send_events(height, events, chain_id.clone())
+                .map_err(Error::worker)?
         }
 
         // If there is a NewBlock event, forward the event to any workers affected by it.
         if let Some(IbcEvent::NewBlock(new_block)) = collected.new_block {
             for worker in self.workers.to_notify(&src_chain.id()) {
-                worker.send_new_block(height, new_block)?;
+                worker
+                    .send_new_block(height, new_block)
+                    .map_err(Error::worker)?
             }
         }
 
         Ok(())
     }
+
+    fn clear_pending_packets(&mut self, chain_id: &ChainId) -> Result<(), Error> {
+        for worker in self.workers.workers_for_chain(chain_id) {
+            worker.clear_pending_packets().map_err(Error::worker)?;
+        }
+
+        Ok(())
+    }
 }
 
 /// Describes the result of [`collect_events`].
