@@ -2,10 +2,11 @@ mod block_on;
 pub use block_on::block_on;
 
 mod recv_multiple;
-pub use recv_multiple::{recv_multiple, try_recv_multiple};
+pub use recv_multiple::try_recv_multiple;
 
 pub mod diff;
 pub mod iter;
+pub mod queue;
 pub mod retry;
 pub mod sled;
 pub mod stream;
