@@ -1,34 +1,34 @@
 use std::{thread, time::Duration};
 
-use anomaly::BoxError;
 use crossbeam_channel::Receiver;
 use tracing::{debug, info, trace, warn};
 
 use ibc::{events::IbcEvent, ics02_client::events::UpdateClient};
 
 use crate::{
-    chain::handle::ChainHandlePair,
-    foreign_client::{ForeignClient, ForeignClientError, MisbehaviourResults},
+    chain::handle::{ChainHandle, ChainHandlePair},
+    foreign_client::{ForeignClient, ForeignClientErrorDetail, MisbehaviourResults},
     object::Client,
     telemetry,
     telemetry::Telemetry,
 };
 
+use super::error::RunError;
 use super::WorkerCmd;
 
-pub struct ClientWorker {
+pub struct ClientWorker<ChainA: ChainHandle, ChainB: ChainHandle> {
     client: Client,
-    chains: ChainHandlePair,
+    chains: ChainHandlePair<ChainA, ChainB>,
     cmd_rx: Receiver<WorkerCmd>,
 
     #[allow(dead_code)]
     telemetry: Telemetry,
 }
 
-impl ClientWorker {
+impl<ChainA: ChainHandle, ChainB: ChainHandle> ClientWorker<ChainA, ChainB> {
     pub fn new(
         client: Client,
-        chains: ChainHandlePair,
+        chains: ChainHandlePair<ChainA, ChainB>,
         cmd_rx: Receiver<WorkerCmd>,
         telemetry: Telemetry,
     ) -> Self {
@@ -41,7 +41,7 @@ impl ClientWorker {
     }
 
     /// Run the event loop for events associated with a [`Client`].
-    pub fn run(self) -> Result<(), BoxError> {
+    pub fn run(self) -> Result<(), RunError> {
         let mut client = ForeignClient::restore(
             self.client.dst_client_id.clone(),
             self.chains.b.clone(),
@@ -70,12 +70,14 @@ impl ClientWorker {
                         )
                     };
                 }
-                Err(e @ ForeignClientError::ExpiredOrFrozen(..)) => {
-                    warn!("[{}] failed to refresh client: {}", client, e);
+                Err(e) => {
+                    if let ForeignClientErrorDetail::ExpiredOrFrozen(_) = e.detail() {
+                        warn!("failed to refresh client '{}': {}", client, e);
 
-                    // This worker has completed its job as the client cannot be refreshed any
-                    // further, and can therefore exit without an error.
-                    return Ok(());
+                        // This worker has completed its job as the client cannot be refreshed any
+                        // further, and can therefore exit without an error.
+                        return Ok(());
+                    }
                 }
                 _ => (),
             };
@@ -95,7 +97,7 @@ impl ClientWorker {
         Ok(())
     }
 
-    fn process_cmd(&self, cmd: WorkerCmd, client: &ForeignClient) -> Next {
+    fn process_cmd(&self, cmd: WorkerCmd, client: &ForeignClient<ChainB, ChainA>) -> Next {
         match cmd {
             WorkerCmd::IbcEvents { batch } => {
                 trace!("[{}] worker received batch: {:?}", client, batch);
@@ -106,7 +108,7 @@ impl ClientWorker {
 
                         // Run misbehaviour. If evidence submitted the loop will exit in next
                         // iteration with frozen client
-                        if self.detect_misbehaviour(&client, Some(update)) {
+                        if self.detect_misbehaviour(client, Some(update)) {
                             telemetry! {
                                 self.telemetry.ibc_client_misbehaviour(
                                     &self.client.dst_chain_id,
@@ -120,12 +122,19 @@ impl ClientWorker {
 
                 Next::Continue
             }
-            WorkerCmd::Shutdown => Next::Abort,
+
             WorkerCmd::NewBlock { .. } => Next::Continue,
+            WorkerCmd::ClearPendingPackets => Next::Continue,
+
+            WorkerCmd::Shutdown => Next::Abort,
         }
     }
 
-    fn detect_misbehaviour(&self, client: &ForeignClient, update: Option<UpdateClient>) -> bool {
+    fn detect_misbehaviour(
+        &self,
+        client: &ForeignClient<ChainB, ChainA>,
+        update: Option<UpdateClient>,
+    ) -> bool {
         match client.detect_misbehaviour_and_submit_evidence(update) {
             MisbehaviourResults::ValidClient => false,
             MisbehaviourResults::VerificationError => {
@@ -145,7 +154,7 @@ impl ClientWorker {
     }
 
     /// Get a reference to the client worker's chains.
-    pub fn chains(&self) -> &ChainHandlePair {
+    pub fn chains(&self) -> &ChainHandlePair<ChainA, ChainB> {
         &self.chains
     }
 
